#include <stddef.h>
#include <string.h>
#include <stdarg.h>
#include <alloc.h>
#include <io.h>
#include <fcntl.h>
#include <sys/stat.h>

#include "modsetup.h"
#include "xscommon.h"
#include "keyboard.h"

#include "setevent.h"

#pragma pack(push, 1)
typedef struct {
  short keys[3];
  short mask;
  char *item;
} def_item;

typedef struct {
  DWORD mark;
  DWORD size;
} iff_item;
#pragma pack(pop)

#define SWAP32(x) ((((x)>>24)&0xFFUL) | (((x)<<8)&0xFF0000UL) | (((x)>>8)&0xFF00UL) | (((x)<<24)&0xFF000000UL))

#define IFF_FORM 0x4D524F46UL
#define IFF_EVNT 0x544E5645UL
#define IFF_INFO 0x4F464E49UL
#define IFF_EVEN 0x4E455645UL

/* moving forward: Up, KPad8 */
#define EVT_MOVE_FORWARD    0
/* moving backward: Down, KPad2 */
#define EVT_MOVE_BACKWARD   1
/* turning left: Left, KPad4 */
#define EVT_TURN_LEFT       2
/* turning right: Right, KPad6 */
#define EVT_TURN_RIGHT      3
/* call menu, go back: Escape */
#define EVT_MENU_BACK       4
/* open, use: LCtrl, RCtrl */
#define EVT_OPEN            5
/* run mode: LAlt, RAlt */
#define EVT_RUN             8
/* toggle HUD: I */
#define EVT_HUD            12
/* hold to show coords (cheat): C */
#define EVT_CHT_COORDS      9
/* hold to walk through walls (cheat): W */
#define EVT_CHT_WALL_WALK  10
/* toggle map: M */
#define EVT_MAP            11
/* toggle infinite health (cheat): H */
#define EVT_CHT_HEALTH     13
/* toggle visibility for enemies (cheat): V */
#define EVT_CHT_VISIBLE    14
/* toggle infinite ammo (cheat): A */
#define EVT_CHT_AMMO       15
/* toggle open full map (cheat): O */
#define EVT_CHT_OPEN_MAP   16
/* keys for weapons and save/load slots: 1..8 */
#define EVT_KEY_1          17
#define EVT_KEY_2          18
#define EVT_KEY_3          19
#define EVT_KEY_4          20
#define EVT_KEY_5          21
#define EVT_KEY_6          22
#define EVT_KEY_7          23
#define EVT_KEY_8          24
/* weapon reload: Space */
#define EVT_RELOAD         25
/* moving with mouse: Left Mouse Button */
#define EVT_MOUSE_CONTROL   6
/* fire weapon: LShift, RShift, Right Mouse Button */
#define EVT_FIRE            7
/* confirm, enter submenu */
#define EVT_CONFIRM_ENTER  26
/* toggle cheats mode (cheat): E + G + J */
#define EVT_CHT_MASTER_KEY 27

/* max msk */
static unsigned char deflists[28][2] = {
  {2, EVT_MOVE_FORWARD},
  {2, EVT_MOVE_BACKWARD},
  {2, EVT_TURN_LEFT},
  {2, EVT_TURN_RIGHT},
  {1, EVT_MENU_BACK},
  {2, EVT_OPEN},
  {2, EVT_RUN},
  {1, EVT_HUD},
  {1, EVT_CHT_COORDS},
  {1, EVT_CHT_WALL_WALK},
  {1, EVT_MAP},
  {1, EVT_CHT_HEALTH},
  {1, EVT_CHT_VISIBLE},
  {1, EVT_CHT_AMMO},
  {1, EVT_CHT_OPEN_MAP},
  {1, EVT_KEY_1},
  {1, EVT_KEY_2},
  {1, EVT_KEY_3},
  {1, EVT_KEY_4},
  {1, EVT_KEY_5},
  {1, EVT_KEY_6},
  {1, EVT_KEY_7},
  {1, EVT_KEY_8},
  {1, EVT_RELOAD},
  {1, EVT_MOUSE_CONTROL},
  {3, EVT_FIRE},
  {1, EVT_CONFIRM_ENTER},
  {1, EVT_CHT_MASTER_KEY}
};

/* default EVENTS.IFF values */
/*
  key1 + key2 + key3 or mouse_button
  -1 - not used
*/
static short defevent[36][4] = {
  /* 0 (1) */
  {0x0148, -1, -1, -1},
  {0x0048, -1, -1, -1},
  /* 1 (2) */
  {0x0150, -1, -1, -1},
  {0x0050, -1, -1, -1},
  /* 2 (4) */
  {0x014B, -1, -1, -1},
  {0x004B, -1, -1, -1},
  /* 3 (8) */
  {0x014D, -1, -1, -1},
  {0x004D, -1, -1, -1},
  /* 4 (10) */
  {0x0001, -1, -1, -1},
  /* 5 (20) */
  {0x001D, -1, -1, -1},
  {0x011D, -1, -1, -1},
  /* 8 (100) */
  {0x0038, -1, -1, -1},
  {0x0138, -1, -1, -1},
  /* 12 (1000) */
  {0x0017, -1, -1, -1},
  /* 9 (200) */
  {0x002E, -1, -1, -1},
  /* 10 (400) */
  {0x0011, -1, -1, -1},
  /* 11 (800) */
  {0x0032, -1, -1, -1},
  /* 13 (2000) */
  {0x0023, -1, -1, -1},
  /* 14 (4000) */
  {0x002F, -1, -1, -1},
  /* 15 (8000) */
  {0x001E, -1, -1, -1},
  /* 16 (10000) */
  {0x0018, -1, -1, -1},
  /* 17 (20000) */
  {0x0002, -1, -1, -1},
  /* 18 (40000) */
  {0x0003, -1, -1, -1},
  /* 19 (80000) */
  {0x0004, -1, -1, -1},
  /* 20 (100000) */
  {0x0005, -1, -1, -1},
  /* 21 (200000) */
  {0x0006, -1, -1, -1},
  /* 22 (400000) */
  {0x0007, -1, -1, -1},
  /* 23 (800000) */
  {0x0008, -1, -1, -1},
  /* 24 (1000000) */
  {0x0009, -1, -1, -1},
  /* 25 (2000000) */
  {0x0039, -1, -1, -1},
  /* 6 (40) */
  {    -1, -1, -1,  1},
  /* 7 (80) */
  {0x002A, -1, -1, -1},
  {0x0036, -1, -1, -1},
  {    -1, -1, -1,  2},
  /* 26 (4000000) */
  {0x001C, -1, -1, -1},
  /* 27 (8000000) */
  {0x0012, 0x0022, 0x0024, -1}
};

static char filename[] = "DATA\\EVENTS.IFF";

def_item gamectrl[] = {
  /* scan1, scan2, mouse, event (1<<value, i.e. bit number), name */
  {{-1, -1, -1}, EVT_MOVE_FORWARD, "Move_Forward"},
  {{-1, -1, -1}, EVT_MOVE_BACKWARD, "Move_Backward"},
  {{-1, -1, -1}, EVT_TURN_LEFT, "Turn_Left"},
  {{-1, -1, -1}, EVT_TURN_RIGHT, "Turn_Right"},
  {{-1, -1, -1}, EVT_FIRE, "Fire"},
  {{-1, -1, -1}, EVT_OPEN, "Open"},
  {{-1, -1, -1}, EVT_RUN, "Run"},
  {{-1, -1, -1}, EVT_RELOAD, "Reload"},
  {{-1, -1, -1}, EVT_MAP, "Map"},
  {{-1, -1, -1}, EVT_HUD, "HUD"},
  {{-1, -1, -1}, EVT_MOUSE_CONTROL, "Mouse_Control"}
};

/*
 1 - Left Mouse Button   MLeft
 2 - Right Mouse Button  MRight
*/

menuitem mousebtn[] = {
  {"Choose mouse button", NULL, 2},
  {"MLeft", NULL, 0},
  {"MRight", NULL, 0}
};

/* v1.1 */
static char smainctrl[(35 + 1) * 11];
menuitem mainctrl[1 + 11];

static char strHint[] = "Press \x1A or \x1B to choose a key for selected event and Delete to remove definition.";

/* v1.1 */
menuitem mainevts[] = {
  {"Configuration setup", "Current Filename ( DATA\\EVENTS.IFF )", 5},
  {"Configure control events", "Configure game keyboard and mouse controls.", 0},
  {"Reset control events", "Reset all control events to thier default values.", 0},
  {"", NULL, ITEM_GRAYED},
  {"Save changes to file", "Save changes to the configuration file.", 0},
  {"Do not save any changes", "Return without saving anything.", 0}
};

/* v1.1 */
menuitem notfound[] = {
  {"Error on loading file", NULL, 1 | MENU_NFOBOX},
  {"Events file \"DATA\\EVENTS.IFF\" invalid or not found!", NULL, 0}
};

/* v1.1 */
menuitem resetevt[] = {
  {"Confirm reset?", NULL, 2 | MENU_DLGBOX},
  {"Yes", NULL, 0},
  {"No", NULL, 0}
};

/* v1.1 */
static menuitem menusave[] = {
  {"Save changes in configuration?", NULL, 2 | MENU_DLGBOX},
  {"Yes", "Save all changes to file.", 0},
  {"No",  "Do not save any changes.", 0}
};

int GetKeyProc(int item, BYTE *key) {
int r;
  r = 0;
  if ((!item) && (!key)) {
    r = GetScanCode();
  }
  return(r);
}

void DrawCols(int cols) {
int i;
  cols %= 3;
  for (i = 1; i <= (mainctrl[0].flag & MENU_ITEMSK); i++) {
    mainctrl[i].flag = cols;
  }
}

char *GetMouseByButton(short k) {
char *s;
  if ((k >= 1) && (k <= 2)) {
    s = mousebtn[k].item;
  } else {
    s = GetNameByScanCode(-1);
  }
  return(s);
}

short CheckKeyEvent(short k, int item, int n) {
short i, j, p, r;
  r = 0;
  if (k != -1) {
    for (i = 1; i <= (mainctrl[0].flag & MENU_ITEMSK); i++) {
      /* keyboard */
      if (n < 2) {
        if (i == item) {
          /* same twice */
          if (
            ((n == 0) && (gamectrl[i - 1].keys[1] == k)) ||
            ((n == 1) && (gamectrl[i - 1].keys[0] == k))
          ) {
            r = i;
            break;
          }
        } else {
          /* duplicate */
          if ((gamectrl[i - 1].keys[1] == k) || (gamectrl[i - 1].keys[0] == k)) {
            r = i;
            break;
          }
        }
      } else {
        /* mouse */
        /* duplicate */
        if ((i != item) && (gamectrl[i - 1].keys[2] == k)) {
          r = i;
          break;
        }
      }
    }
    /* reserved keys - keyboard only */
    if (n < 2) {
      p = 0;
      for (i = 0; i < 28; i++) {
        for (j = 0; j < 28; j++) {
          if (deflists[i][1] == gamectrl[j].mask) {
            break;
          }
        }
        /* not found */
        if (j >= 28) {
          for (j = 0; j < deflists[i][0]; j++) {
            /* system key used */
            if (
              (defevent[p + j][0] == k) ||
              (defevent[p + j][1] == k) ||
              (defevent[p + j][2] == k)
            ) {
              r = -1;
              break;
            }
          }
        }
        p += deflists[i][0];
      }
    }
  }
  return(r);
}

void SetMenuItem(menuitem *mitem, char *item, char *hint, WORD flag) {
  if (mitem) {
    mitem->item = item;
    mitem->hint = hint;
    mitem->flag = flag;
  }
}

int CtrlProc(int item, BYTE *key) {
menuitem infotext;
char *s, b[60];
short k, t;
  if (key) {
    switch (*key) {
      case VK_DELETE:
        s = GetNameByScanCode(-1);
        gamectrl[item - 1].keys[mainctrl[item].flag] = -1;
        memset(&mainctrl[item].item[15 + (7 * mainctrl[item].flag)], ' ', 6);
        memcpy(&mainctrl[item].item[15 + (7 * mainctrl[item].flag)], s, strlen(s));
        break;
      case VK_LEFT:
        mainctrl[item].flag -= (mainctrl[item].flag > 0) ? 1 : (-2);
        break;
      case VK_TAB:
      case VK_RIGHT:
        mainctrl[item].flag += (mainctrl[item].flag < 2) ? 1 : (-2);
        break;
    }
    DrawCols(mainctrl[item].flag);
  } else {
    do {
      /* default key */
      k = gamectrl[item - 1].keys[mainctrl[item].flag];
      if (mainctrl[item].flag < 2) {
        /* keyboard */
        SetMenuItem(&infotext, "Press a Key", NULL, 0);
        t = DrawInfo(&infotext, GetKeyProc);
        /* Esc - do not change anything */
        if (t != 1) { k = t; }
        s = GetNameByScanCode(k);
        /* do not change anything */
        if (t == 1) { break; }
      } else {
        /* mouse */
        t = DrawMenu(mousebtn, NULL);
        /* Esc - do not change anything */
        if (t != 0) { k = t; }
        s = GetMouseByButton(k);
        /* do not change anything */
        if (t == 0) { break; }
      }
      t = CheckKeyEvent(k, item, mainctrl[item].flag);
      if (t) {
        if (t == item) {
          StrMerge(b, 60, "ERROR: ", s, " cannot be assigned to ", gamectrl[item - 1].item, " twice.", NULL);
          k = -1;
        } else {
          if (t == -1) {
            StrMerge(b, 60, "WARNING: ", s, " already in use as reserved game key.", NULL);
          } else {
            StrMerge(b, 60, "WARNING: ", s, " has already been assigned to ", gamectrl[t - 1].item, ".", NULL);
          }
        }
        SetMenuItem(&infotext, b, NULL, 1);
        DrawInfo(&infotext, NULL);
      }
    } while (k <= 0);
    gamectrl[item - 1].keys[mainctrl[item].flag] = k;
    memset(&mainctrl[item].item[15 + (7 * mainctrl[item].flag)], ' ', 6);
    memcpy(&mainctrl[item].item[15 + (7 * mainctrl[item].flag)], s, strlen(s));
  }
  return(1);
}

void LoadDefaults(void) {
int i, j, k, p;
  /* default config */
  p = 0;
  for (i = 0; i < 28; i++) {
    for (j = 0; j < 11; j++) {
      if (deflists[i][1] == gamectrl[j].mask) {
        /* drop any defined keys */
        for (k = 0; k < 3; k++) {
          gamectrl[j].keys[k] = -1;
        }
        /* set defaults */
        for (k = 0; k < deflists[i][0]; k++) {
          if (defevent[p + k][0] != -1) {
            gamectrl[j].keys[k ? 1 : 0] = defevent[p + k][0];
          }
          if (defevent[p + k][3] != -1) {
            gamectrl[j].keys[2] = defevent[p + k][3];
          }
        }
        break;
      }
    }
    p += deflists[i][0];
  }
}

void BuildItems(void) {
int i, j;
char *s;
  /* v1.1 reduce executable file size */
  memset(mainctrl, 0, sizeof(mainctrl));
  mainctrl[0].item = "Choose a key definition to modify";
  mainctrl[0].hint = "Event          Key1   Key2   Mouse ";
  mainctrl[0].flag = (LIST_LEN(mainctrl) - 1) | MENU_ROWCOL;
  /* initialize with controls strings */
  for (i = 1; i <= (mainctrl[0].flag & MENU_ITEMSK); i++) {
    mainctrl[i].flag = ITEM_HILITE; /* for FillMenu() */
    mainctrl[i].item = gamectrl[i - 1].item;
  }
  /* "EventTitleText KeyDef KeyDef MBttn " */
  FillMenu(mainctrl, smainctrl, 35);
  for (i = 1; i <= (mainctrl[0].flag & MENU_ITEMSK); i++) {
    for (j = 0; j < 3; j++) {
      if (j < 2) {
        s = GetNameByScanCode(gamectrl[i - 1].keys[j]);
      } else {
        s = GetMouseByButton(gamectrl[i - 1].keys[j]);
      }
      memcpy(&mainctrl[i].item[15 + (j * 7)], s, strlen(s));
    }
    mainctrl[i].hint = strHint;
  }
  DrawCols(0);
}

iff_item *NextChunk(iff_item *item, int bdeep) {
DWORD sz;
  if (item) {
    sz = sizeof(item[0]);
    sz += bdeep ? ((item->mark == IFF_FORM) ? 4 : 0) : SWAP32(item->size);
    /* padding for odd chunk size */
    sz += sz & 1;
    item = (iff_item *) (((BYTE *) item) + ((WORD) sz));
  }
  return(item);
}

int CheckChunks(BYTE *p, DWORD fsz) {
int result, c;
iff_item *item;
DWORD sz;
  c = 0;
  result = 0;
  item = (iff_item *) p;
  if (
    (p) && (fsz >= sizeof(item[0]) + 4) && (item->mark == IFF_FORM) &&
    (fsz >= (SWAP32(item->size) + sizeof(item[0]))) && (*((DWORD *)&p[sizeof(item[0])]) == IFF_EVNT)
  ) {
    fsz = SWAP32(item->size) - 4;
    p += (sizeof(item[0]) + 4);
    while (fsz >= sizeof(item[0])) {
      item = (iff_item *) p;
      sz = SWAP32(item->size);
      if (fsz >= (sz + sizeof(item[0]))) {
        /* check inside container */
        if (item->mark == IFF_FORM) {
          result = CheckChunks(p, fsz);
        } else {
          /* valid chunks */
          if (
            ((item->mark == IFF_INFO) && (sz == 4)) ||
            ((item->mark == IFF_EVEN) && (sz == 12)) ||
            ((item->mark == IFF_EVEN) && (sz == 14))
          ) {
            result = 1;
          }
        }
        sz += sizeof(item[0]);
        /* padding for odd chunk size */
        sz += (sz & 1);
        if (fsz >= sz) {
          fsz -= sz;
          p += (WORD) sz;
        } else {
          result = 0;
        }
      }
      if (!result) { break; }
      c++;
      if (!fsz) { break; }
      result = 0;
    }
  }
  return(result ? c : 0);
}

int LoadEvents(void) {
iff_item *form, *item;
int fl, i, j, k, c, u;
DWORD sz, *d;
BYTE *p;
  c = 0;
  fl = open(filename, O_RDONLY | O_BINARY);
  if (fl != -1) {
    lseek(fl, 0, SEEK_END);
    sz = tell(fl);
    p = (BYTE *) malloc((WORD) sz);
    if (p) {
      lseek(fl, 0, SEEK_SET);
      read(fl, p, (WORD) sz);
      /* check chunks */
      c = CheckChunks(p, sz);
      if (c) {
        /* FORM */
        form = NextChunk((iff_item *) p, 1);
        for (i = 0; i < c; i++) {
          /* FORM/FORM */
          if (form->mark == IFF_FORM) {
            u = CheckChunks((BYTE *) form, SWAP32(form->size) + sizeof(form[0]));
            if (u) {
              item = NextChunk(form, 1);
              /* FORM/FORM/INFO */
              if (item->mark == IFF_INFO) {
                d = (DWORD *) NextChunk(item, 1);
                for (k = 0; k < 11; k++) {
                  if (*d == (((DWORD)1) << gamectrl[k].mask)) {
                    break;
                  }
                }
                /* event found */
                if (k < 11) {
                  /* drop defaults */
                  for (j = 0; j < 3; j++) {
                    gamectrl[k].keys[j] = -2;
                  }
                  for (j = 0; j < (u - 1); j++) {
                    item = NextChunk(item, 0);
                    if (item->mark == IFF_EVEN) {
                      d = (DWORD *) NextChunk(item, 1);
                      if (SWAP32(item->size) == 12) {
                        /* keyboard */
                        gamectrl[k].keys[(gamectrl[k].keys[0] == -2) ? 0 : 1] = *d;
                      } else {
                        /* mouse */
                        gamectrl[k].keys[2] = *((WORD *) &d[3]);
                      }
                    }
                  }
                  /* v1.1 fix unassigned keys */
                  for (j = 0; j < 3; j++) {
                    if (gamectrl[k].keys[j] < -1) {
                      gamectrl[k].keys[j] = -1;
                    }
                  }
                }
              }
            }
          }
          form = NextChunk(form, 0);
        }
      }
      free(p);
    }
    close(fl);
  }
  return(c);
}

void SaveEvents(void) {
int fl, i, j, k, p;
DWORD sz, d[4], bs;
iff_item item;
  fl = open(filename, O_RDWR | O_BINARY | O_CREAT | O_TRUNC, S_IREAD | S_IWRITE);
  if (fl != -1) {
    /* FORM */
    item.mark = IFF_FORM;
    item.size = 0;
    write(fl, &item, sizeof(item));
    item.mark = IFF_EVNT;
    write(fl, &item, 4);
    sz = 4;
    /* write data */
    p = 0;
    for (i = 0; i < 28; i++) {
      /* FORM/FORM */
      sz += sizeof(item) + 4;
      bs = tell(fl);
      item.mark = IFF_FORM;
      item.size = 0;
      write(fl, &item, sizeof(item));
      item.mark = IFF_EVNT;
      write(fl, &item, 4);
      /* FORM/FORM/INFO */
      sz += sizeof(item) + 4;
      item.mark = IFF_INFO;
      item.size = SWAP32(4UL);
      write(fl, &item, sizeof(item));
      item.mark = 1;
      item.mark <<= deflists[i][1];
      write(fl, &item, 4);
      /* FORM/FORM/EVEN */
      for (j = 0; j < 11; j++) {
        if (deflists[i][1] == gamectrl[j].mask) {
          break;
        }
      }
      /* v1.1 - do not assign default keys */
      /*/- no keys set - use default -/
      if (
        (gamectrl[j].keys[0] == -1) &&
        (gamectrl[j].keys[1] == -1) &&
        (gamectrl[j].keys[2] == -1)
      ) {
        j = 11;
      }*/
      if (j < 11) {
        /* found */
        for (k = 0; k < 3; k++) {
          d[0] = gamectrl[j].keys[k];
          d[1] = (DWORD) -1;
          d[2] = (DWORD) -1;
          d[3] = (DWORD) -1;
          if (d[0] != ((DWORD) -1)) {
            item.mark = IFF_EVEN;
            if (k < 2) {
              /* keyboard */
              item.size = 12;
            } else {
              /* mouse */
              item.size = 14;
              d[3] = d[0];
              d[0] = (DWORD) -1;
            }
            item.mark = IFF_EVEN;
            sz += item.size + sizeof(item);
            item.size = SWAP32(item.size);
            write(fl, &item, sizeof(item));
            write(fl, d, (WORD) SWAP32(item.size));
          }
        }
      } else {
        /* use default */
        for (j = 0; j < deflists[i][0]; j++) {
          d[0] = defevent[p + j][0];
          d[1] = defevent[p + j][1];
          d[2] = defevent[p + j][2];
          d[3] = defevent[p + j][3];
          item.mark = IFF_EVEN;
          item.size = (d[3] == ((DWORD) -1)) ? 12 : 14;
          sz += item.size + sizeof(item);
          item.size = SWAP32(item.size);
          write(fl, &item, sizeof(item));
          write(fl, d, (WORD) SWAP32(item.size));
        }
      }
      p += deflists[i][0];
      /* fix FORM/FORM size */
      item.mark = IFF_FORM;
      item.size = tell(fl) - bs - sizeof(item);
      item.size = SWAP32(item.size);
      lseek(fl, bs, SEEK_SET);
      write(fl, &item, sizeof(item));
      lseek(fl, 0, SEEK_END);
    }
    /* fix FORM size */
    lseek(fl, 0, SEEK_SET);
    item.mark = IFF_FORM;
    item.size = SWAP32(sz);
    write(fl, &item, sizeof(item));
    close(fl);
  }
}

/* v1.1 from this point and below */
int LoadEventsData(void) {
int r;
  /* load default keys */
  LoadDefaults();
  /* try to load events from file */
  r = LoadEvents();
  if (!r) {
    DrawMenu(notfound, NULL);
  }
  return(r);
}

int MenuEvts(int item, BYTE *key) {
  if (!key) {
    switch (item) {
      /* change */
      case 1:
        BuildItems();
        DrawMenu(mainctrl, CtrlProc);
        break;
      /* reset */
      case 2:
        if (DrawMenu(resetevt, NULL) == 1) {
          LoadDefaults();
        }
        break;
      /* exit */
      default:
        item = 0;
        break;
    }
  }
  return(item);
}

void DoEventsMenu(void) {
int i;
  if (LoadEventsData()) {
    i = DrawMenu(mainevts, MenuEvts);
    /* save if needed */
    if (
      /* save selected */
      ((i + 1) == (mainevts[0].flag & MENU_ITEMSK)) ||
      /* or ask for saving changes */
      ((!i) && (DrawMenu(menusave, NULL) == 1))
    ) {
      SaveEvents();
    }
  }
}
